<?php
/**
 * USAGE:
 * $pos = new BSXPrinter($host, $port, $password);
 * $receipt = new BSXReceipt();
 * $receipt->addItem('Osadnicy z Catanu', 99, 1, 23)
 *         ->addItem('Dobble', 59.95, 2, 23);
 * $pos->printReceipt($receipt, 500);
 * echo "PRINTED: ".$pos->getLastDoc();

Osadnicy z Catanu       1x99.00   99.00B
Dobble                 2x59.95   119.90B
- - - - - - - - - - - - - - - - - - - - 
Sprzed.opodatk.B                  218.90
Kwota PTU B 23.00 %                40.93
CZNA KWOTA PTU                   40.93
SUMA                              218.90
- - - - - - - - - - - - - - - - - - - - 
          N I E F I S K A L N Y
      #kasa      kasjer            14:37
- - - - - - - - - - - - - - - - - - - - 
Gotwka                           500.00
Reszta                            281.10

 * @author Artur Jedlinski <artur@jedlinski.pl>
 * @copyright 2014 XPECT.pl
 * @version 0.1 @ 2014-06-10
 * @license http://www.gnu.org/copyleft/lesser.html GNU Lesser General Public License 
  */
class BSXPrinter
{
    const LINE_START = "\x01";
    const LINE_END = "\r\n";
    const SEPARATOR = '#';
    
    public $verbose = false;
    private $socket;
    protected $response = null;
    public $maxDelay = 10;

    /**
     * @param string $host
     * @param int $port
     * @param string $password
     * @param bool $verbose
     * @throws RuntimeException
     */
    public function __construct($host = null, $port = 7361, $password = null, $verbose = false)
    {
        $this->socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        if ($this->socket === false) {
            throw new RuntimeException("socket_create() failed, reason: " . socket_strerror(socket_last_error()));
        }
        
        $this->verbose = $verbose;
        if (!is_null($host)) {
            $this->connect($host, (int)$port, $password);
        }
    }
    
    public function __destruct()
    {
        socket_close($this->socket);
    }
    
    /**
     * @param string $host
     * @param int $port
     * @param string $password
     * @throws RuntimeException
     */
    public function connect($host, $port = 7361, $password = null)
    {
        $result = socket_connect($this->socket, $host, $port);
        if ($result === false) {
            throw new RuntimeException("socket_connect() failed.\nReason: ($result) " . socket_strerror(socket_last_error($socket)));
        }
        
        if (!empty($password)) {
            $this->sendCmd('PASSWORD', [ $password ]);
        }
        
        return $this;
    }

    /**
     * 
     * @param string $line
     * @return BSXPrinter
     * @throws RuntimeException
     * @throws BSXPrinterException
     */
    public function sendLine($line)
    {
        $this->log($line);
        
        $in = self::LINE_START.$line.self::LINE_END;
        socket_write($this->socket, $in, strlen($in));
        
        $out = null; $ticks = 0;
        while (empty($out) && ($ticks < $this->maxDelay))
        {
            $ticks++;
            $out = socket_read($this->socket, 1024, PHP_NORMAL_READ);
            $out = trim($out, self::LINE_START.self::LINE_END);
            if (strlen($out) == 0) {
                usleep(100000);
            }
        }
        
        if (strlen($out) == 0) {
            throw new RuntimeException("connection timeout");
        }
        
        $this->log($out);
        if ((substr($out, 0, 1) != '+')) {
            throw new RuntimeException("invalid server response: ".$out);
        }
        
        $this->response = explode(self::SEPARATOR, substr($out, 1));
        if ($this->response[0] == 'ERR') {
            $error = substr($out, 5);
            throw new BSXPrinterException($out);
        }
        
        if ($this->response[0] != 'OK') {
            throw new RuntimeException("invalid server response: ".$out);
        }
        
        return $this;
    }
    
    /**
     * 
     * @param string $cmd
     * @param string|array $params
     * @throws InvalidArgumentException
     */
    public function sendCmd($cmd, $params = null)
    {
        $line = self::SEPARATOR.$cmd;
        if (!empty($params)) {
            if (is_string($params)) {
                $line .= self::SEPARATOR.$params;
            }
            elseif (is_numeric($params)) {
                $line .= self::SEPARATOR.$params;
            }
            elseif (is_array($params)) {
                $line .= self::SEPARATOR.join(self::SEPARATOR, $params);
            }
            else {
                throw new InvalidArgumentException(sprintf("BSXPrinter::sendCmd() accepts string or array as second parameter, %s given.",
                    is_object($params) ? get_class($params) : gettype($params)));
            }
        }
        
        $this->sendLine($line);
        return $this;
    }
    
    /**
     * 
     * @param array|BSXReceipt $items
     * @return BSXPrinter
     */
    public function printReceipt($items, $gotowka = null, $karta = null)
    {
        $this->sendCmd('RECEIPT');
        if ($items instanceof BSXReceipt) {
            $items = $items->items;
        }
        
        $total = 0;
        foreach ($items as $item) {
            $this->sendCmd('ITEM', [
                $item['name'],
                round($item['price'], 2),
                $item['amount'],
                round($item['vat'], 2),
                round($item['price'] * $item['amount'], 2),
            ]);
            
            $total += $item['price'] * $item['amount'];
        }
        
        $this->sendCmd('COMMIT', [ round($total, 2), $gotowka, $karta ]);
        return $this;
    }
    
    public function getVersion()
    {
        $this->sendCmd('GETVERSION');
        return $this->response[1];
    }
    
    public function getProtocol()
    {
        $this->sendCmd('GETPROTOCOL');
        return $this->response[1];
    }
    
    /**
     * @return BSXPrinter
     */
    public function reportCash()
    {
        $this->sendCmd('REPORTCASH');
        return $this;
    }
    
    /**
     * @return BSXPrinter
     */
    public function reportDay()
    {
        $this->sendCmd('REPORTDAY');
        return $this;
    }
    
    /**
     * @return array
     */
    public function getResponse()
    {
        return $this->response;
    }
    
    /**
     * @return string
     */
    public function getLastDoc()
    {
        $this->sendCmd('GETLASTDOC');
        return $this->response[1];
    }
    
    /**
     * @return string
     */
    public function getLastDocId()
    {
        $this->sendCmd('GETLASTDOCID');
        return $this->response[1];
    }
    
    /**
     * @return float
     */
    public function getCash()
    {
        $this->sendCmd('CASH');
        return $this->response[1];
    }
    
    /**
     * @param string $txt
     * @return BSXPrinter
     */
    protected function log($txt)
    {
        if ($this->verbose) {
            echo $txt."\n<br />";
        }
        
        return $this;
    }
}

class BSXReceipt
{
    public $items = [];
    
    public function addItem($name, $price, $amount, $vat)
    {
        $this->items[] = [
            'name' => $name,
            'price' => $price,
            'amount' => (int)$amount,
            'vat' => $vat
            
        ];
        
        return $this;
    }
}

class BSXPrinterException extends RuntimeException
{
    
}